from typing import List, Optional

from pythonbible.books import Book
from pythonbible.normalized_reference import NormalizedReference
from pythonbible.verses import (
    MAX_VERSE_NUMBER_BY_BOOK_AND_CHAPTER,
    VERSE_IDS,
    get_number_of_verses,
    get_verse_id,
)


def is_valid_verse_id(verse_id: int) -> bool:
    """
    Checks to see if the given verse_id corresponds to a valid verse in the Bible

    :param verse_id: a verse id
    :type verse_id: int
    :return: True if the verse_id is in the list of valid verse ids; otherwise, False
    :rtype: bool
    """
    return verse_id in VERSE_IDS


def is_valid_reference(reference: NormalizedReference) -> bool:
    """
    Checks to see if the given NormalizedReference is a valid scripture reference
    (i.e. all of the verses in the reference are valid verses)

    :param reference: a normalized reference
    :type reference: NormalizedReference
    :return: True if the reference is valid; otherwise, False
    :rtype: bool
    """
    if reference is None or not isinstance(reference, NormalizedReference):
        return False

    if not is_valid_verse(
        reference.book, reference.start_chapter, reference.start_verse
    ):
        return False

    if not is_valid_verse(reference.book, reference.end_chapter, reference.end_verse):
        return False

    start_verse_id: int = get_verse_id(
        reference.book, reference.start_chapter, reference.start_verse
    )
    end_verse_id: int = get_verse_id(
        reference.book, reference.end_chapter, reference.end_verse
    )

    return start_verse_id <= end_verse_id


def is_valid_book(book: Book) -> bool:
    """
    Checks to see if the given book is a valid book of the Bible

    :param book: a book of the Bible
    :type book: Book
    :return: True if the given book is valid; otherwise, False
    :rtype: bool
    """
    return book is not None and isinstance(book, Book)


def is_valid_chapter(book: Book, chapter: int) -> bool:
    """
    Checks to see if the given Book is a valid book of the Bible; and, if so, checks
    to see if the given chapter number is a valid chapter number for the given book

    :param book: a book of the Bible
    :type book: Book
    :param chapter: a chapter number for the given book of the Bible
    :type chapter: int
    :return: True if the given book and chapter are valid; otherwise, False
    :rtype: bool
    """
    if not is_valid_book(book):
        return False

    if chapter is None or not isinstance(chapter, int):
        return False

    chapter_list: Optional[List[int]] = MAX_VERSE_NUMBER_BY_BOOK_AND_CHAPTER.get(book)

    return chapter_list is not None and 1 <= chapter <= len(chapter_list)


def is_valid_verse(book: Book, chapter: int, verse: int) -> bool:
    """
    Checks to see if the given :ref:`Book` is a valid book of the Bible,
    then checks to see if the given chapter number is a valid chapter number for the given book,
    then checks to see if the given verse number is a valid verse number for the given book and chapter

    :param book: a book of the Bible
    :type book: :ref:`Book`
    :param chapter: a chapter number for the given book of the Bible
    :type chapter: int
    :param verse: a verse number for the given book and chapter
    :type verse: int
    :return: True if the given book, chapter, and verse are valid; otherwise, False
    :rtype: bool
    """
    if not is_valid_chapter(book, chapter):
        return False

    if verse is None or not isinstance(verse, int):
        return False

    max_verse: int = get_number_of_verses(book, chapter)

    return 1 <= verse <= max_verse



import pytest

import pythonbible as bible
from pythonbible.bible.osis.constants import get_book_by_id


def test_get_book_by_id() -> None:
    assert get_book_by_id("Gen") == bible.Book.GENESIS


def test_get_book_by_id_null() -> None:
    with pytest.raises(bible.InvalidBookError):
        get_book_by_id("blah")



from functools import singledispatch
from typing import List, Optional

from pythonbible.books import Book
from pythonbible.normalized_reference import NormalizedReference
from pythonbible.parser import get_references
from pythonbible.verses import get_number_of_chapters, get_number_of_verses


@singledispatch
def count_verses(references: List[NormalizedReference]) -> int:
    """
    Returns the count of verses included in the given list of references.

    :param references: A list of normalized references
    :type references: List[NormalizedReference]
    :return: The count of verses included in the given list of references
    :rtype: int
    """
    return _get_number_verses_in_references(references)


@count_verses.register
def _count_verses_single(reference: NormalizedReference) -> int:
    return _get_number_of_verses_in_reference(reference)


@count_verses.register
def _count_verses_string(reference: str) -> int:
    return _get_number_verses_in_references(get_references(reference))


def _get_number_verses_in_references(references: List[NormalizedReference]) -> int:
    return sum(
        _get_number_of_verses_in_reference(reference) for reference in references
    )


def _get_number_of_verses_in_reference(reference: NormalizedReference) -> int:
    number_of_verses: int = 0
    start_book = reference.book
    end_book = reference.end_book or start_book

    for book_id in range(start_book.value, end_book.value + 1):
        book: Book = Book(book_id)
        start_chapter: int = reference.start_chapter if book == start_book else 1
        end_chapter: int = (
            reference.end_chapter if book == end_book else get_number_of_chapters(book)
        )

        for chapter in range(start_chapter, end_chapter + 1):
            start_verse: Optional[int] = (
                reference.start_verse
                if book == start_book and chapter == reference.start_chapter
                else None
            )
            end_verse: Optional[int] = (
                reference.end_verse
                if book == end_book and chapter == reference.end_chapter
                else None
            )

            number_of_verses += _get_number_of_verses_in_chapter(
                book, chapter, start_verse, end_verse
            )

    return number_of_verses


def _get_number_of_verses_in_chapter(
    book: Book, chapter: int, start_verse: Optional[int], end_verse: Optional[int]
) -> int:
    return (end_verse or get_number_of_verses(book, chapter)) - (start_verse or 1) + 1



from functools import singledispatch
from typing import List

from pythonbible.books import Book
from pythonbible.normalized_reference import NormalizedReference
from pythonbible.parser import get_references
from pythonbible.verses import get_number_of_chapters


@singledispatch
def count_chapters(references: List[NormalizedReference]) -> int:
    """
    Returns the count of chapters of books the Bible included in the given list of references.

    :param references: A list of normalized references
    :type references: List[NormalizedReference]
    :return: The count of chapters of books of the Bible included in the given list of references
    :rtype: int
    """
    return _get_number_of_chapters_in_references(references)


@count_chapters.register
def _count_chapters_single(reference: NormalizedReference) -> int:
    return _get_number_of_chapters_in_reference(reference)


@count_chapters.register
def _count_chapters_string(reference: str) -> int:
    return _get_number_of_chapters_in_references(get_references(reference))


def _get_number_of_chapters_in_references(references: List[NormalizedReference]) -> int:
    return sum(
        _get_number_of_chapters_in_reference(reference) for reference in references
    )


def _get_number_of_chapters_in_reference(reference: NormalizedReference) -> int:
    if not reference.end_book or reference.book == reference.end_book:
        return reference.end_chapter - reference.start_chapter + 1

    # Start book chapters
    number_of_chapters: int = (
        get_number_of_chapters(reference.book) - reference.start_chapter + 1
    )

    # Middle book(s) chapters
    for book_id in range(reference.book.value + 1, reference.end_book.value):
        number_of_chapters += get_number_of_chapters(Book(book_id))

    # End book chapters
    number_of_chapters += reference.end_chapter

    return number_of_chapters



from typing import Any, Optional


class InvalidBookError(Exception):
    """Raised when the book id is not valid."""


class InvalidChapterError(Exception):
    """Raised when the chapter number is not a valid chapter number for the given book of the Bible."""


class InvalidVerseError(Exception):
    """Raised when the verse id or book, chapter, and verse number being processed is not a valid Bible verse."""

    def __init__(self, message: Optional[str] = None, **kwargs) -> None:
        """

        :param message: optional message string
        :param kwargs: optional keyword arguments (verse_id, book, chapter, verse) for more specific messaging
        """
        self.message: Optional[str] = message
        self.verse_id: Optional[Any] = kwargs.get("verse_id")
        self.book: Optional[Any] = kwargs.get("book")
        self.chapter: Optional[Any] = kwargs.get("chapter")
        self.verse: Optional[Any] = kwargs.get("verse")

        if not self.message:
            if self.book and self.chapter and self.verse:
                self.message = f"{self.book.title} {self.chapter}:{self.verse} is not a valid verse."
            elif self.verse_id:
                self.message = f"{self.verse_id} is not a valid verse."

        super(InvalidVerseError, self).__init__(self.message)


class InvalidBibleParserError(Exception):
    """Raised when the Bible parser is not valid."""


class MissingVerseFileError(Exception):
    """Raised when the verse file for a given version is not found."""


class MissingBookFileError(Exception):
    """Raised when the book file for a given version is not found."""
